/* ========================================
 *
 * Copyright CONSULTRON, 2021
 * All Rights Reserved
 * UNPUBLISHED, LICENSED SOFTWARE.
 *
 * CONSULTRON allows PUBLIC use of the file.
 * CONSULTRON provides NO WARRANTY expressed or implied.
 *
 * ========================================
*/
/*------------------------------------------------
Goal
This project is a replacement of the standard string library functions:
•	memset()
•	memcpy()
•	memmove()
with DMA-driven equivalent versions of the functions for the purpose of decreasing the transfer time.
•	memset() =>	dmemset()
•	memcpy() => dmemcpy()
•	memmove() => dmemmove()
The demo code in main() compares the standard functions to the DMA equivalents with framing pulses to allow an
oscilloscope to measure the timing differences at different transfer sizes.

Theory of Operation
A single DMA channel and a single TD are allocated to accomplish this task.  
When each of the DMA functions are called the Channel and related TD are configured for the memory pointers and size.

Note:  The DMA-equivalent functions are limited to 4095 bytes of transfer.  
Also, small transfer sizes (below 64 bytes) are not faster than the standard functions.  
At BUS_CLK = 72MHz, a DMA transfer of 4095 bytes is at least 7 times faster than the standard functions.
 * ========================================
*/
#include "project.h"
#include <stdio.h>

uint32_t systick = 0;
uint32_t systick_prev = 0;
CY_ISR(isr_systick)
{
	systick++;
}

#define VDAC_STEP_TIME 1000	// time in msec

/* Defines for DMA_memmove */
#define DMA_memmove_BYTES_PER_BURST 0		/* Let the TD size determine the burst size */
#define DMA_memmove_REQUEST_PER_BURST 1		/* The DMA channel stops after the burst is complete */
#define DMA_memmove_SRC_BASE (CYDEV_SRAM_BASE)
#define DMA_memmove_DST_BASE (CYDEV_SRAM_BASE)
#define DMA_MAX_TD_TRANSFER_SZ 	4095		/* Maximum HW allowed TD transfer size */

/* DMA memmove error codes */
#define DMA_memmove_NO_ERROR					0
#define DMA_memmove_INVALID_TRANSFERCOUNT		-1
#define DMA_memmove_MEMTYPE_INVALID				-2
#define DMA_memmove_NOT_REQUESTED				-3
#define DMA_memmove_DMA_BUSY					-4

/* Variable declarations for DMA_memmove */
/* Move these variable declarations to the top of the function */
uint8 DMA_memmove_Chan;
uint8 DMA_memmove_TD[1];


/*******************************************************************************
* Function Name: DMA_memmove_Init
****************************************************************************
* DMA Configuration for DMA_memmove 
*
* \return void
*/
void DMA_memmove_Init(void)
{
	DMA_memmove_Chan = DMA_memmove_DmaInitialize(DMA_memmove_BYTES_PER_BURST, DMA_memmove_REQUEST_PER_BURST, 
	    HI16(DMA_memmove_SRC_BASE), HI16(DMA_memmove_DST_BASE));
	DMA_memmove_TD[0] = CyDmaTdAllocate();
	/* At this point the TD is not initialized with the transfercount, nextTD, configuration or the src or dst addresses */
	CyDmaChSetInitialTd(DMA_memmove_Chan, DMA_memmove_TD[0]);
	CyDmaChEnable(DMA_memmove_Chan, 1);
}

int DMA_memmove_err;
_Bool DMA_memmove_rqstd = 0;


/*******************************************************************************
* Function Name: dmemmove_error
****************************************************************************
* DMA memmove error return function
*
* \return DMA_memmove_NO_ERROR if no error occurred.
* \return The error code set by the DMA memmove functions.
*/
int8 dmemmove_error( void )
{
	return(DMA_memmove_err);
}


/*******************************************************************************
* Function Name: dmemmove_cmplt
****************************************************************************
* DMA memmove wait for transfer completion function.  (Blocking function)
* It tests if a request was made and if so, if the transfer is completed.
*
* \return The DMA transfer is complete (DMA_memmove_NO_ERROR)
* \return The error code set by the DMA memmove functions.
*/
int8 dmemmove_cmplt( void )
{
uint8 dma_stat;
	/* Check to make sure the DMA was requested */
	if(DMA_memmove_rqstd == 0) {DMA_memmove_err = DMA_memmove_NOT_REQUESTED;}
	else
	{
		DMA_memmove_rqstd = 0;	/* the request was made.  Clear this flag */

		/* Check to make sure the DMA requested is completed */
		while(CyDmaChGetRequest(DMA_memmove_Chan) != 0) {} 	/* First wait for the request to be accepted */
		do 	{ 
			CyDmaChStatus(DMA_memmove_Chan, NULL, &dma_stat);	/* Get the DMA channel's status */
		}while((dma_stat & CY_DMA_STATUS_TD_ACTIVE) == CY_DMA_STATUS_TD_ACTIVE);	/* loop if still active */
		/* DMA completed */
	}
	return(DMA_memmove_err);
}


/*******************************************************************************
* Function Name: dmemset_rqst
****************************************************************************
* DMA memmove set REQUEST function.  (Non-Blocking function)
* It makes a DMA request to fill the dest memory with the fill byte.
* It doesn't wait for the transfer to complete.
*
* \param dest Pointer to the destination memory.
*
* \param value Byte value to set (fill) the dest.
*
* \param num Size of dest transfer.  NOTE: This parameter is limited to DMA_MAX_TD_TRANSFER_SZ
*
* \return Pointers to the dest memory
* \return NULL.  An error occurred.  Check \ref dmemmove_err() for the error code.
*
* \sideeffect
*  There is NO testing for either pointers + transfer_sz as valid for project resources.
*  It is up to the designer to make sure of the proper parameters to 
*  prevent accidental memory overwriting which can result in data corruption.
*/
void *dmemset_rqst( void * dest, int value, size_t num )
{
uint8 fill[4];
uint8 dma_stat;
/* check that the DMA is not busy.  This is needed because we are using a shared resource. */
	CyDmaChStatus(DMA_memmove_Chan, NULL, &dma_stat);
	if(dma_stat & CY_DMA_STATUS_TD_ACTIVE)	/* check that the DMA is busy */
		{DMA_memmove_err = DMA_memmove_DMA_BUSY; return NULL;}	/* busy */
		
/* check that the size to transfer is valid.  DMA_MAX_TD_TRANSFER_SZ is the max allowed size. */
	if((num == 0) || (num > DMA_MAX_TD_TRANSFER_SZ)) 
		{DMA_memmove_err = DMA_memmove_INVALID_TRANSFERCOUNT; return NULL;}	/* transfer size too big */
	DMA_memmove_err = 0;

	fill[0] = fill[1] = fill[2] = fill[3] = (0xFF & value); /* only use the LSB of value and fill the long with the value */
	
	CyDmaChSetExtendedAddress(DMA_memmove_Chan, HI16(CYDEV_SRAM_BASE), HI16(DMA_memmove_DST_BASE));
	CyDmaTdSetConfiguration(DMA_memmove_TD[0], num, DMA_memmove_TD[0], CY_DMA_TD_INC_DST_ADR | DMA_memmove__TD_TERMOUT_EN);
	CyDmaTdSetAddress(DMA_memmove_TD[0], LO16((uint32)fill), LO16((uint32)dest));
	CyDmaChSetRequest(DMA_memmove_Chan,CY_DMA_CPU_REQ);	/* Request the DMA to execute */
	DMA_memmove_rqstd = 1;
	return(dest);
}

/*******************************************************************************
* Function Name: dmemset
****************************************************************************
* DMA memmove set function.  Equivalent to memset(). (Blocking function)
* It makes a DMA request to fill the dest memory with the fill byte
* and WAITS for the transfer to complete.
*
* \param dest Pointer to the destination memory.
*
* \param value Byte value to set (fill) the dest.
*
* \param num Size of dest transfer.  NOTE: This parameter is limited to DMA_MAX_TD_TRANSFER_SZ
*
* \return Pointers to the dest memory (if successful)
* \return NULL.  An error occurred.  Check \ref dmemmove_err() for the error code.
*
* \sideeffect
*  There is NO testing for either pointers + transfer_sz as valid for project resources.
*  It is up to the designer to make sure of the proper parameters to 
*  prevent accidental memory overwriting which can result in data corruption.
*/
void *dmemset( void * dest, int value, size_t num )
{
void *result;
	if((result = dmemset_rqst(dest, value, num)) != NULL)
	{
		dmemmove_cmplt();	/* Return when DMA complete */
	}
	return(result);
}


/*******************************************************************************
* Function Name: dmemcpy_rqst
****************************************************************************
* DMA memmove copy REQUEST function.  (Non-Blocking function)
* It makes a DMA request to transfer the src memory the dest memory.
* It doesn't wait for the transfer to complete.
*
* This function will adjust the src base address based on the src memory pointer and size to transfer.
* Only CYDEV_SRAM_BASE and CYDEV_FLASH_BASE are acceptable at this time.
*
* \param dest Pointer to the destination memory.
*
* \param src Pointer to the source memory.
*
* \param num Size of dest transfer.  NOTE: This parameter is limited to DMA_MAX_TD_TRANSFER_SZ
*
* \return Pointers to the dest memory
* \return NULL.  An error occurred.  Check \ref dmemmove_err() for the error code.
*
* \sideeffect
*  There is NO testing for either pointers + transfer_sz as valid for project resources.
*  It is up to the designer to make sure of the proper parameters to 
*  prevent accidental memory overwriting which can result in data corruption.
*/
void *dmemcpy_rqst(void *dest, const void * src, size_t num)
{
uint8 dma_stat;
/* check that the DMA is not busy.  This is needed because we are using a shared resource. */
	CyDmaChStatus(DMA_memmove_Chan, NULL, &dma_stat);
	if(dma_stat & CY_DMA_STATUS_TD_ACTIVE)	/* check that the DMA is busy */
		{DMA_memmove_err = DMA_memmove_DMA_BUSY; return NULL;}	/* busy */
		
/* check that the size to transfer is valid.  DMA_MAX_TD_TRANSFER_SZ is the max allowed size. */
	if((num == 0) || (num > DMA_MAX_TD_TRANSFER_SZ)) 
		{DMA_memmove_err = DMA_memmove_INVALID_TRANSFERCOUNT; return NULL;}	/* transfer size too big */
	DMA_memmove_err = 0;
	
	/* Adjust source base for source memory type */
	if((src >= (void *)CYDEV_SRAM_BASE ) && ((src + num) < (void *)CYDEV_SRAM_BASE + CYDEV_SRAM_SIZE ))
	{	/* Src = SRAM */
		CyDmaChSetExtendedAddress(DMA_memmove_Chan, HI16(CYDEV_SRAM_BASE), HI16(DMA_memmove_DST_BASE));
	}
	else if((src >= (void *)CYDEV_FLASH_BASE ) && ((src + num) < (void *)CYDEV_SRAM_BASE + CYDEV_FLASH_SIZE ))
	{	/* Src = FLASH */
		CyDmaChSetExtendedAddress(DMA_memmove_Chan, HI16(CYDEV_FLASH_BASE), HI16(DMA_memmove_DST_BASE));
	}
	else {DMA_memmove_err = DMA_memmove_MEMTYPE_INVALID;  return NULL;}	/* memory base not allowed */
	
	CyDmaTdSetConfiguration(DMA_memmove_TD[0], num, DMA_memmove_TD[0], CY_DMA_TD_INC_SRC_ADR | CY_DMA_TD_INC_DST_ADR | DMA_memmove__TD_TERMOUT_EN);
	CyDmaTdSetAddress(DMA_memmove_TD[0], LO16((uint32)src), LO16((uint32)dest));
	CyDmaChSetRequest(DMA_memmove_Chan,CY_DMA_CPU_REQ);	/* Request the DMA to execute */
	DMA_memmove_rqstd = 1;
	return(dest);
}


/*******************************************************************************
* Function Name: dmemcpy
****************************************************************************
* DMA memmove copy function.  Equivalent to memcpy(). (Blocking function)
* It makes a DMA request to transfer the src memory the dest memory
* and WAITS for the transfer to complete.
*
* \param dest Pointer to the destination memory.
*
* \param src Pointer to the source memory.
*
* \param num Size of dest transfer.  NOTE: This parameter is limited to DMA_MAX_TD_TRANSFER_SZ
*
* \return Pointers to the dest memory
* \return NULL.  An error occurred.  Check \ref dmemmove_err() for the error code.
*
* \sideeffect
*  There is NO testing for either pointers + transfer_sz as valid for project resources.
*  It is up to the designer to make sure of the proper parameters to 
*  prevent accidental memory overwriting which can result in data corruption.
*/
void *dmemcpy(void *dest, const void * src, size_t num)
{
void *result;
	if((result = dmemcpy_rqst(dest, src, num)) != NULL)
	{
		dmemmove_cmplt();	/* Return when DMA complete */
	}
	return(result);
}
#define dmemmove  dmemcpy	/* Equivalent function to memove(). */


#define SZ_XFER	DMA_MAX_TD_TRANSFER_SZ
const uint8_t mem_src[SZ_XFER] = {1,128,245,65,189,45,92,198,34,246,70,55,127,255,[DMA_MAX_TD_TRANSFER_SZ-3]=2,[DMA_MAX_TD_TRANSFER_SZ-2]=219,[DMA_MAX_TD_TRANSFER_SZ-1]=45};
uint8_t mem_dst[SZ_XFER];
uint16 xfer_sz_tbl[] = {1,2,4,8,16,32,64,128,256,512,1024,2048,4095};

/*************************************************************************************/
int main(void)
{
    CyGlobalIntEnable; /* Enable global interrupts. */

	Term_Start();
	CySysTickStart();
	CySysTickSetCallback(0, isr_systick);

	DMA_memmove_Init();

    for(;;)
    {
		if(systick >= (systick_prev + VDAC_STEP_TIME))
		{
			systick_prev = systick;

/* Get baseline timing reading the framing pulse.  Timing to be subtracted from later acquired timings */
Pin_code_Write(1);	/* Framing pulse start */
Pin_code_Write(0);	/* Framing pulse end */

/* Step through the transfer sizes table */
			for(uint16 i=0; i<sizeof(xfer_sz_tbl)/sizeof(xfer_sz_tbl[0]); i++)
			{
/*** Let's try the standard memcpy() and memset() functions ***/
Pin_code_Write(1);	/* Framing pulse start */
				memset(mem_dst,0x55,xfer_sz_tbl[i]);
Pin_code_Write(0);	/* Framing pulse end */
Pin_code_Write(1);	/* Framing pulse start */
				memcpy(mem_dst,mem_src,xfer_sz_tbl[i]);
Pin_code_Write(0);	/* Framing pulse end */			
/***/

/*** Let's try the DMA-based dmemcpy() and dmemset() functions ***/
Pin_code_Write(1);	/* Framing pulse start */
				dmemset(mem_dst,0x88,xfer_sz_tbl[i]);
Pin_code_Write(0);	/* Framing pulse end */
Pin_code_Write(1);	/* Framing pulse start */
				dmemcpy(mem_dst, mem_src, xfer_sz_tbl[i]);
Pin_code_Write(0);	/* Framing pulse end */
/***/
			}
		}
    }
}


/* [] END OF FILE */
